@@ -5,7 +5,7 @@ use cranelift_codegen::entity::SecondaryMap;
 use cranelift_codegen::isa::TargetIsa;
 use cranelift_codegen::{self, ir};
 use cranelift_codegen::{
-    binemit::{Addend, CodeInfo, CodeOffset, Reloc, RelocSink, StackMapSink, TrapSink},
+    binemit::{Addend, CodeOffset, Reloc, RelocSink, StackMapSink, TrapSink},
     CodegenError,
 };
 use cranelift_module::{
@@ -114,7 +114,7 @@ impl ObjectBuilder {
 /// See the `ObjectBuilder` for a convenient way to construct `ObjectModule` instances.
 pub struct ObjectModule {
     isa: Box<dyn TargetIsa>,
-    object: Object,
+    object: Object<'static>,
     declarations: ModuleDeclarations,
     functions: SecondaryMap<FuncId, Option<(SymbolId, bool)>>,
     data_objects: SecondaryMap<DataId, Option<(SymbolId, bool)>>,
@@ -310,27 +310,17 @@ impl Module for ObjectModule {
         trap_sink: &mut dyn TrapSink,
         stack_map_sink: &mut dyn StackMapSink,
     ) -> ModuleResult<ModuleCompiledFunction> {
-        info!(
-            "defining function {}: {}",
-            func_id,
-            ctx.func.display(self.isa())
-        );
-        let CodeInfo {
-            total_size: code_size,
-            ..
-        } = ctx.compile(self.isa())?;
-        let mut code: Vec<u8> = vec![0; code_size as usize];
+        info!("defining function {}: {}", func_id, ctx.func.display());
+        let mut code: Vec<u8> = Vec::new();
         let mut reloc_sink = ObjectRelocSink::default();
 
-        unsafe {
-            ctx.emit_to_memory(
-                &*self.isa,
-                code.as_mut_ptr(),
-                &mut reloc_sink,
-                trap_sink,
-                stack_map_sink,
-            )
-        };
+        ctx.compile_and_emit(
+            self.isa(),
+            &mut code,
+            &mut reloc_sink,
+            trap_sink,
+            stack_map_sink,
+        )?;
 
         self.define_function_bytes(func_id, &code, &reloc_sink.relocs)
     }
@@ -585,6 +575,11 @@ impl ObjectModule {
                 32,
             ),
             Reloc::X86GOTPCRel4 => (RelocationKind::GotRelative, RelocationEncoding::Generic, 32),
+            Reloc::Arm64Call => (
+                RelocationKind::Relative,
+                RelocationEncoding::AArch64Call,
+                26,
+            ),
             Reloc::ElfX86_64TlsGd => {
                 assert_eq!(
                     self.object.format(),
@@ -613,8 +608,32 @@ impl ObjectModule {
                     32,
                 )
             }
+            Reloc::Aarch64TlsGdAdrPage21 => {
+                assert_eq!(
+                    self.object.format(),
+                    object::BinaryFormat::Elf,
+                    "Aarch64TlsGdAdrPrel21 is not supported for this file format"
+                );
+                (
+                    RelocationKind::Elf(object::elf::R_AARCH64_TLSGD_ADR_PAGE21),
+                    RelocationEncoding::Generic,
+                    21,
+                )
+            }
+            Reloc::Aarch64TlsGdAddLo12Nc => {
+                assert_eq!(
+                    self.object.format(),
+                    object::BinaryFormat::Elf,
+                    "Aarch64TlsGdAddLo12Nc is not supported for this file format"
+                );
+                (
+                    RelocationKind::Elf(object::elf::R_AARCH64_TLSGD_ADD_LO12_NC),
+                    RelocationEncoding::Generic,
+                    12,
+                )
+            }
             // FIXME
-            _ => unimplemented!(),
+            reloc => unimplemented!("{:?}", reloc),
         };
         ObjectRelocRecord {
             offset: record.offset,
@@ -645,7 +664,7 @@ fn translate_linkage(linkage: Linkage) -> (SymbolScope, bool) {
 /// compilation.
 pub struct ObjectProduct {
     /// Object artifact with all functions and data from the module defined.
-    pub object: Object,
+    pub object: Object<'static>,
     /// Symbol IDs for functions (both declared and defined).
     pub functions: SecondaryMap<FuncId, Option<(SymbolId, bool)>>,
     /// Symbol IDs for data objects (both declared and defined).
@@ -710,28 +729,4 @@ impl RelocSink for ObjectRelocSink {
             name: name.clone(),
         })
     }
-
-    fn reloc_jt(&mut self, _offset: CodeOffset, reloc: Reloc, _jt: ir::JumpTable) {
-        match reloc {
-            Reloc::X86PCRelRodata4 => {
-                // Not necessary to record this unless we are going to split apart code and its
-                // jumptbl/rodata.
-            }
-            _ => {
-                panic!("Unhandled reloc");
-            }
-        }
-    }
-
-    fn reloc_constant(&mut self, _offset: CodeOffset, reloc: Reloc, _jt: ir::ConstantOffset) {
-        match reloc {
-            Reloc::X86PCRelRodata4 => {
-                // Not necessary to record this unless we are going to split apart code and its
-                // jumptbl/rodata.
-            }
-            _ => {
-                panic!("Unhandled reloc");
-            }
-        }
-    }
 }
